import { createHooks } from 'hookable';
import { tagDedupeKey, sortCriticalTags, resolveTitleTemplateFromTags, HasElementTags, ValidHeadTags, normaliseTag as normaliseTag$1 } from 'zhead';

let activeHead;
const setActiveHead = (head) => activeHead = head;
const getActiveHead = () => activeHead;

function defineHeadPlugin(plugin) {
  return plugin;
}

const DedupesTagsPlugin = defineHeadPlugin({
  hooks: {
    "tag:normalise": function({ tag }) {
      ["hid", "vmid", "key"].forEach((key) => {
        if (tag.props[key]) {
          tag.key = tag.props[key];
          delete tag.props[key];
        }
      });
      const dedupe = tag.key ? `${tag.tag}:${tag.key}` : tagDedupeKey(tag);
      if (dedupe)
        tag._d = dedupe;
    },
    "tags:resolve": function(ctx) {
      const deduping = {};
      ctx.tags.forEach((tag, i) => {
        let dedupeKey = tag._d || tag._p || i;
        const dupedTag = deduping[dedupeKey];
        if (dupedTag) {
          let strategy = tag?.tagDuplicateStrategy;
          if (!strategy && (tag.tag === "htmlAttrs" || tag.tag === "bodyAttrs"))
            strategy = "merge";
          if (strategy === "merge") {
            const oldProps = dupedTag.props;
            ["class", "style"].forEach((key) => {
              if (tag.props[key] && oldProps[key])
                tag.props[key] = `${oldProps[key]} ${tag.props[key]}`;
            });
            deduping[dedupeKey].props = {
              ...oldProps,
              ...tag.props
            };
            return;
          } else if (tag._e === dupedTag._e) {
            dedupeKey = `${dedupeKey}:entry(${tag._e}:${tag._p})`;
            tag._d = dedupeKey;
          } else {
            tag._p = dupedTag._p;
          }
          if (Object.keys(tag.props).length === 0 && !tag.children) {
            delete deduping[dedupeKey];
            return;
          }
        }
        deduping[dedupeKey] = tag;
      });
      ctx.tags = Object.values(deduping);
    }
  }
});

const SortTagsPlugin = defineHeadPlugin({
  hooks: {
    "tags:resolve": (ctx) => {
      const tagIndexForKey = (key) => ctx.tags.find((tag) => tag._d === key)?._p;
      for (const tag of ctx.tags) {
        if (!tag?.tagPriority)
          continue;
        if (typeof tag.tagPriority === "number") {
          tag._p = tag.tagPriority;
          continue;
        }
        const modifiers = [{ prefix: "before:", offset: -1 }, { prefix: "after:", offset: 1 }];
        for (const { prefix, offset } of modifiers) {
          if (tag.tagPriority.startsWith(prefix)) {
            const key = tag.tagPriority.replace(prefix, "");
            const index = tagIndexForKey(key);
            if (typeof index !== "undefined")
              tag._p = index + offset;
          }
        }
      }
      ctx.tags.sort((a, b) => a._p - b._p).sort(sortCriticalTags);
    }
  }
});

const TitleTemplatePlugin = defineHeadPlugin({
  hooks: {
    "tags:resolve": (ctx) => {
      ctx.tags = resolveTitleTemplateFromTags(ctx.tags);
    }
  }
});

function asArray(value) {
  return Array.isArray(value) ? value : [value];
}
const TagConfigKeys = ["tagPosition", "tagPriority", "tagDuplicateStrategy"];
function hashCode(s) {
  let h = 9;
  for (let i = 0; i < s.length; )
    h = Math.imul(h ^ s.charCodeAt(i++), 9 ** 9);
  return ((h ^ h >>> 9) + 65536).toString(16).substring(1, 7).toLowerCase();
}

const HydratesStatePlugin = defineHeadPlugin({
  hooks: {
    "tag:normalise": (ctx) => {
      const { tag, entry } = ctx;
      if (!HasElementTags.includes(tag.tag))
        return;
      if (typeof tag._d === "undefined" && entry._m === "server")
        return;
      const hasChildren = tag.children && tag.children.length > 0;
      tag._s = `data-h-${hashCode(tag._d || tag.tag + (hasChildren ? tag.children : JSON.stringify(tag.props)))}`;
      tag.props[tag._s] = "";
    }
  }
});

function normaliseTag(tagName, input, entry) {
  const tag = normaliseTag$1(tagName, input, { childrenKeys: ["innerHTML", "textContent"] });
  tag._e = entry._i;
  Object.keys(tag.props).filter((k) => TagConfigKeys.includes(k)).forEach((k) => {
    tag[k] = tag.props[k];
    delete tag.props[k];
  });
  if (typeof tag.props.class === "object" && !Array.isArray(tag.props.class)) {
    tag.props.class = Object.keys(tag.props.class).filter((k) => tag.props.class[k]);
  }
  if (Array.isArray(tag.props.class))
    tag.props.class = tag.props.class.join(" ");
  if (tag.props.content && Array.isArray(tag.props.content)) {
    return tag.props.content.map((v, i) => {
      const newTag = { ...tag, props: { ...tag.props } };
      newTag.props.content = v;
      newTag.key = `${tag.props.name || tag.props.property}:${i}`;
      return newTag;
    });
  }
  return tag;
}
function normaliseEntryTags(e) {
  return Object.entries(e.input).filter(([k, v]) => typeof v !== "undefined" && ValidHeadTags.includes(k)).map(
    ([k, value]) => asArray(value).map((props) => asArray(normaliseTag(k, props, e)))
  ).flat(3).map((t, i) => {
    t._p = (e._i << 8) + i++;
    return t;
  });
}

function createHead(options = {}) {
  let entries = [];
  let _sde = {};
  let entryId = 0;
  const hooks = createHooks();
  if (options.hooks)
    hooks.addHooks(options.hooks);
  const plugins = [
    DedupesTagsPlugin,
    SortTagsPlugin,
    TitleTemplatePlugin
  ];
  plugins.push(...options.plugins || []);
  plugins.forEach((plugin) => hooks.addHooks(plugin.hooks || {}));
  const head = {
    _removeQueuedSideEffect(key) {
      delete _sde[key];
    },
    _flushQueuedSideEffects() {
      Object.values(_sde).forEach((fn) => fn());
      _sde = {};
    },
    headEntries() {
      return entries;
    },
    get hooks() {
      return hooks;
    },
    push(input, options2) {
      const _i = entryId++;
      entries.push({
        _i,
        input,
        _sde: {},
        ...options2
      });
      return {
        dispose() {
          entries = entries.filter((e) => {
            if (e._i !== _i)
              return true;
            _sde = { ..._sde, ...e._sde || {} };
            e._sde = {};
            return false;
          });
        },
        patch(input2) {
          entries = entries.map((e) => {
            if (e._i === _i) {
              _sde = { ..._sde, ...e._sde || {} };
              e._sde = {};
              e.input = e._i === _i ? input2 : e.input;
            }
            return e;
          });
        }
      };
    },
    async resolveTags() {
      const resolveCtx = { tags: [], entries: [...entries] };
      await hooks.callHook("entries:resolve", resolveCtx);
      for (const entry of resolveCtx.entries) {
        for (const tag of normaliseEntryTags(entry)) {
          const tagCtx = { tag, entry };
          await hooks.callHook("tag:normalise", tagCtx);
          resolveCtx.tags.push(tagCtx.tag);
        }
      }
      await hooks.callHook("tags:resolve", resolveCtx);
      return resolveCtx.tags;
    }
  };
  setActiveHead(head);
  return head;
}

function useHead(input, options = {}) {
  const head = getActiveHead();
  head.push(input, options);
}
const useTitle = (title) => {
  useHead({ title });
};
const useMeta = (meta) => {
  useHead({ meta: [meta] });
};
const useLink = (link) => {
  useHead({ link: [link] });
};
const useScript = (script) => {
  useHead({ script: [script] });
};
const useStyle = (style) => {
  useHead({ style: [style] });
};
const useBase = (base) => {
  useHead({ base });
};
const useHtmlAttrs = (attrs) => {
  useHead({ htmlAttrs: attrs });
};
const useBodyAttrs = (attrs) => {
  useHead({ bodyAttrs: attrs });
};
const useTitleTemplate = (titleTemplate) => {
  useHead({ titleTemplate });
};
const useNoscript = (noscript) => {
  useHead({ noscript: [noscript] });
};

export { DedupesTagsPlugin, HydratesStatePlugin, SortTagsPlugin, TagConfigKeys, TitleTemplatePlugin, activeHead, asArray, createHead, defineHeadPlugin, getActiveHead, hashCode, setActiveHead, useBase, useBodyAttrs, useHead, useHtmlAttrs, useLink, useMeta, useNoscript, useScript, useStyle, useTitle, useTitleTemplate };
