import { NestedHooks, Hookable } from 'hookable';
import { Head, HeadTag } from '@unhead/schema';

declare type HeadPlugin<O> = Omit<CreateHeadOptions<O>, 'plugins'>;
declare function defineHeadPlugin<O = Head>(plugin: HeadPlugin<O>): HeadPlugin<O>;

interface SSRHeadPayload {
    headTags: string;
    bodyTags: string;
    bodyTagsOpen: string;
    htmlAttrs: string;
    bodyAttrs: string;
}
declare function renderSSRHead<T extends HeadClient<any>>(ctx: T): Promise<SSRHeadPayload>;

/**
 * An active head entry provides an API to manipulate it.
 */
interface ActiveHeadEntry<T> {
    /**
     * Updates the entry with new input.
     *
     * Will first clear any side effects for previous input.
     */
    patch: (resolvedInput: T) => void;
    /**
     * Dispose the entry, removing it from the active head.
     *
     * Will queue side effects for removal.
     */
    dispose: () => void;
}
/**
 * Side effects are mapped with a key and their cleanup function.
 *
 * For example `meta:data-h-4h46h465`: () => { document.querySelector('meta[data-h-4h46h465]').remove() }
 */
declare type SideEffectsRecord = Record<string, () => void>;
declare type RuntimeMode = 'server' | 'client' | 'all';
interface HeadEntry<T> {
    /**
     * User provided input for the entry.
     */
    input: T;
    /**
     * The mode that the entry should be used in.
     *
     * @internal
     */
    _m?: RuntimeMode;
    /**
     * Head entry index
     *
     * @internal
     */
    _i: number;
    /**
     * Side effects
     *
     * @internal
     */
    _sde: SideEffectsRecord;
}
declare type HookResult = Promise<void> | void;
interface DomRenderTagContext {
    head: HeadClient;
    tag: HeadTag;
    document: Document;
}
interface EntryResolveCtx<T> {
    tags: HeadTag[];
    entries: HeadEntry<T>[];
}
interface HeadHooks<T> {
    'entries:resolve': (ctx: EntryResolveCtx<T>) => HookResult;
    'tag:normalise': (ctx: {
        tag: HeadTag;
        entry: HeadEntry<T>;
    }) => HookResult;
    'tags:resolve': (ctx: {
        tags: HeadTag[];
    }) => HookResult;
    'dom:beforeRender': (ctx: {
        head: HeadClient;
        tags: HeadTag[];
        document: Document;
    }) => HookResult;
    'dom:renderTag': (ctx: DomRenderTagContext) => HookResult;
    'ssr:beforeRender': (ctx: {
        tags: HeadTag[];
    }) => HookResult;
    'ssr:render': (ctx: {
        tags: HeadTag[];
        html: SSRHeadPayload;
    }) => HookResult;
}
interface CreateHeadOptions<T> {
    plugins?: HeadPlugin<any>[];
    hooks?: NestedHooks<HeadHooks<T>>;
}
interface HeadEntryOptions {
    mode?: RuntimeMode;
}
interface HeadClient<T = Head> {
    /**
     * The active head entries.
     */
    headEntries: () => HeadEntry<T>[];
    /**
     * Create a new head entry.
     */
    push: (entry: T, options?: HeadEntryOptions) => ActiveHeadEntry<T>;
    /**
     * Resolve tags from head entries.
     */
    resolveTags: () => Promise<HeadTag[]>;
    /**
     * Exposed hooks for easier extension.
     */
    hooks: Hookable<HeadHooks<T>>;
    /**
     * @internal
     */
    _removeQueuedSideEffect: (key: string) => void;
    /**
     * @internal
     */
    _flushQueuedSideEffects: () => void;
}

export { ActiveHeadEntry as A, CreateHeadOptions as C, DomRenderTagContext as D, EntryResolveCtx as E, HeadPlugin as H, RuntimeMode as R, SideEffectsRecord as S, HeadClient as a, HeadEntryOptions as b, HeadEntry as c, defineHeadPlugin as d, HookResult as e, HeadHooks as f, SSRHeadPayload as g, renderSSRHead as r };
